#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ladspa.h"

#include "imp/imp.h"

#define IMP_INPUT                0
#define IMP_OUTPUT               1

LADSPA_Descriptor *impDescriptor = NULL;

typedef struct {
	LADSPA_Data *input;
	LADSPA_Data *output;
	imp_buffer * buffer;
} Imp;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {
	switch (index) {
	case 0:
		return impDescriptor;
	default:
		return NULL;
	}
}

void cleanupImp(LADSPA_Handle instance) {
	free(instance);
}

void connectPortImp(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	Imp *plugin;

	plugin = (Imp *)instance;
	switch (port) {
	case IMP_INPUT:
		plugin->input = data;
		break;
	case IMP_OUTPUT:
		plugin->output = data;
		break;
	}
}

LADSPA_Handle instantiateImp(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	Imp *plugin_data = (Imp *)malloc(sizeof(Imp));
	imp_buffer *buffer = NULL;

	printf("creating buffer...\n");
	                        buffer = create_imp_buffer();
	printf("created buffer\n");

	plugin_data->buffer = buffer;

	return (LADSPA_Handle)plugin_data;
}

void runImp(LADSPA_Handle instance, unsigned long sample_count) {
	Imp *plugin_data = (Imp *)instance;
	unsigned long pos;

	/* Input (array of floats of length sample_count) */
	LADSPA_Data *input = plugin_data->input;

	/* Output (array of floats of length sample_count) */
	LADSPA_Data *output = plugin_data->output;
	imp_buffer * buffer = plugin_data->buffer;

	for (pos = 0; pos < sample_count; pos++) {
	        *(output++) = imp(*(input++), buffer);
	}
}

void _init() {
	char **port_names;
	LADSPA_PortDescriptor *port_descriptors;
	LADSPA_PortRangeHint *port_range_hints;

	impDescriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (impDescriptor) {
		impDescriptor->UniqueID = 1199;
		impDescriptor->Label = strdup("imp");
		impDescriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		impDescriptor->Name =
		 strdup("Imp");
		impDescriptor->Maker =
		 strdup("Steve Harris <steve@totl.net>");
		impDescriptor->Copyright =
		 strdup("GPL");
		impDescriptor->PortCount = 2;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(2,
		 sizeof(LADSPA_PortDescriptor));
		impDescriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(2,
		 sizeof(LADSPA_PortRangeHint));
		impDescriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(2, sizeof(char*));
		impDescriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Input */
		port_descriptors[IMP_INPUT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[IMP_INPUT] =
		 strdup("Input");
		port_range_hints[IMP_INPUT].HintDescriptor = 0;

		/* Parameters for Output */
		port_descriptors[IMP_OUTPUT] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[IMP_OUTPUT] =
		 strdup("Output");
		port_range_hints[IMP_OUTPUT].HintDescriptor = 0;

		impDescriptor->activate = NULL;
		impDescriptor->cleanup = cleanupImp;
		impDescriptor->connect_port = connectPortImp;
		impDescriptor->deactivate = NULL;
		impDescriptor->instantiate = instantiateImp;
		impDescriptor->run = runImp;
		impDescriptor->run_adding = NULL;
		impDescriptor->set_run_adding_gain = NULL;
	}
}

void _fini() {
	int i;

	if (impDescriptor) {
		free((char *)impDescriptor->Label);
		free((char *)impDescriptor->Name);
		free((char *)impDescriptor->Maker);
		free((char *)impDescriptor->Copyright);
		free((LADSPA_PortDescriptor *)impDescriptor->PortDescriptors);
		for (i = 0; i < impDescriptor->PortCount; i++)
			free((char *)(impDescriptor->PortNames[i]));
		free((char **)impDescriptor->PortNames);
		free((LADSPA_PortRangeHint *)impDescriptor->PortRangeHints);
		free(impDescriptor);
	}

}
