#include <stdlib.h>
#include <string.h>

#include "ladspa.h"

#define AMP_INPUT                0
#define AMP_OUTPUT               1

LADSPA_Descriptor *nullDescriptor = NULL;

typedef struct {
	LADSPA_Data *input;
	LADSPA_Data *output;
} Null;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {
	switch (index) {
	case 0:
		return nullDescriptor;
	default:
		return NULL;
	}
}

void cleanupNull(LADSPA_Handle instance) {
	free(instance);
}

void connectPortNull(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	Null *plugin;

	plugin = (Null *)instance;
	switch (port) {
	case AMP_INPUT:
		plugin->input = data;
		break;
	case AMP_OUTPUT:
		plugin->output = data;
		break;
	}
}

LADSPA_Handle instantiateNull(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	return malloc(sizeof(Null));
}

void runNull(LADSPA_Handle instance, unsigned long snullle_count) { }

void _init() {
	char **port_names;
	LADSPA_PortDescriptor *port_descriptors;
	LADSPA_PortRangeHint *port_range_hints;

	nullDescriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (nullDescriptor) {
		nullDescriptor->UniqueID = 1;
		nullDescriptor->Label = strdup("null");
		nullDescriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		nullDescriptor->Name =
		 strdup("Null plugin");
		nullDescriptor->Maker =
		 strdup("Steve Harris <steve@totl.net>");
		nullDescriptor->Copyright =
		 strdup("GPL");
		nullDescriptor->PortCount = 2;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(2,
		 sizeof(LADSPA_PortDescriptor));
		nullDescriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(2,
		 sizeof(LADSPA_PortRangeHint));
		nullDescriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(2, sizeof(char*));
		nullDescriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Input */
		port_descriptors[AMP_INPUT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[AMP_INPUT] =
		 strdup("Input");
		port_range_hints[AMP_INPUT].HintDescriptor = 0;

		/* Parameters for Output */
		port_descriptors[AMP_OUTPUT] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[AMP_OUTPUT] =
		 strdup("Output");
		port_range_hints[AMP_OUTPUT].HintDescriptor = 0;

		nullDescriptor->activate = NULL;
		nullDescriptor->cleanup = cleanupNull;
		nullDescriptor->connect_port = connectPortNull;
		nullDescriptor->deactivate = NULL;
		nullDescriptor->instantiate = instantiateNull;
		nullDescriptor->run = runNull;
		nullDescriptor->run_adding = NULL;
		nullDescriptor->set_run_adding_gain = NULL;
	}
}

void _fini() {
	int i;

	if (nullDescriptor) {
		free((char *)nullDescriptor->Label);
		free((char *)nullDescriptor->Name);
		free((char *)nullDescriptor->Maker);
		free((char *)nullDescriptor->Copyright);
		free((LADSPA_PortDescriptor *)nullDescriptor->PortDescriptors);
		for (i = 0; i < nullDescriptor->PortCount; i++)
			free((char *)(nullDescriptor->PortNames[i]));
		free((char **)nullDescriptor->PortNames);
		free((LADSPA_PortRangeHint *)nullDescriptor->PortRangeHints);
		free(nullDescriptor);
	}

}
