#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ladspa.h"

#define MOVERDRV_LIMIT           0
#define MOVERDRV_DRIVE           1
#define MOVERDRV_LOCOLOUR        2
#define MOVERDRV_HICOLOUR        3
#define MOVERDRV_INPUT           4
#define MOVERDRV_OUTPUT          5
#define SOVERDRV_LIMIT           0
#define SOVERDRV_DRIVE           1
#define SOVERDRV_LOCOLOUR        2
#define SOVERDRV_HICOLOUR        3
#define SOVERDRV_INPUT1          4
#define SOVERDRV_INPUT2          5
#define SOVERDRV_OUTPUT1         6
#define SOVERDRV_OUTPUT2         7

LADSPA_Descriptor *mOverDrvDescriptor = NULL;

typedef struct {
	LADSPA_Data *limit;
	LADSPA_Data *drive;
	LADSPA_Data *locolour;
	LADSPA_Data *hicolour;
	LADSPA_Data *input;
	LADSPA_Data *output;
} MOverDrv;

LADSPA_Descriptor *sOverDrvDescriptor = NULL;

typedef struct {
	LADSPA_Data *limit;
	LADSPA_Data *drive;
	LADSPA_Data *locolour;
	LADSPA_Data *hicolour;
	LADSPA_Data *input1;
	LADSPA_Data *input2;
	LADSPA_Data *output1;
	LADSPA_Data *output2;
} SOverDrv;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {
	switch (index) {
	case 0:
		return mOverDrvDescriptor;
	case 1:
		return sOverDrvDescriptor;
	default:
		return NULL;
	}
}

void cleanupMOverDrv(LADSPA_Handle instance) {
	free(instance);
}

void connectPortMOverDrv(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	MOverDrv *plugin;

	plugin = (MOverDrv *)instance;
	switch (port) {
	case MOVERDRV_LIMIT:
		plugin->limit = data;
		break;
	case MOVERDRV_DRIVE:
		plugin->drive = data;
		break;
	case MOVERDRV_LOCOLOUR:
		plugin->locolour = data;
		break;
	case MOVERDRV_HICOLOUR:
		plugin->hicolour = data;
		break;
	case MOVERDRV_INPUT:
		plugin->input = data;
		break;
	case MOVERDRV_OUTPUT:
		plugin->output = data;
		break;
	}
}

LADSPA_Handle instantiateMOverDrv(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	return malloc(sizeof(MOverDrv));
}

void runMOverDrv(LADSPA_Handle instance, unsigned long sample_count) {
	MOverDrv *plugin_data = (MOverDrv *)instance;
	unsigned long pos;

	/* Amps limit (in dB relative to nominal 0) (float value) */
	LADSPA_Data limit = *(plugin_data->limit);

	/* Drive level (float value) */
	LADSPA_Data drive = *(plugin_data->drive);

	/* Low density colouration (float value) */
	LADSPA_Data locolour = *(plugin_data->locolour);

	/* High density colouration (float value) */
	LADSPA_Data hicolour = *(plugin_data->hicolour);

	/* Input (array of floats of length sample_count) */
	LADSPA_Data *input = plugin_data->input;

	/* Output (array of floats of length sample_count) */
	LADSPA_Data *output = plugin_data->output;

	LADSPA_Data gain_fix = pow(10, limit/20);
	
	for (pos = 0; pos < sample_count; pos++) {
	        /* Amplitude of signal */
	        LADSPA_Data amp = *(input++) / gain_fix;
	
	        *(output++) = ((((1.0 - drive) * amp) + (((1.0 / (1.0 + exp(-4.0 * amp))) - 0.5) * 2.0 * drive) + (0.1 * locolour * sin(amp * 23) * amp) + (0.1 * hicolour * cos(amp * 997) * amp))) * gain_fix;
	}
}

void cleanupSOverDrv(LADSPA_Handle instance) {
	free(instance);
}

void connectPortSOverDrv(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	SOverDrv *plugin;

	plugin = (SOverDrv *)instance;
	switch (port) {
	case SOVERDRV_LIMIT:
		plugin->limit = data;
		break;
	case SOVERDRV_DRIVE:
		plugin->drive = data;
		break;
	case SOVERDRV_LOCOLOUR:
		plugin->locolour = data;
		break;
	case SOVERDRV_HICOLOUR:
		plugin->hicolour = data;
		break;
	case SOVERDRV_INPUT1:
		plugin->input1 = data;
		break;
	case SOVERDRV_INPUT2:
		plugin->input2 = data;
		break;
	case SOVERDRV_OUTPUT1:
		plugin->output1 = data;
		break;
	case SOVERDRV_OUTPUT2:
		plugin->output2 = data;
		break;
	}
}

LADSPA_Handle instantiateSOverDrv(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	return malloc(sizeof(SOverDrv));
}

void runSOverDrv(LADSPA_Handle instance, unsigned long sample_count) {
	SOverDrv *plugin_data = (SOverDrv *)instance;
	unsigned long pos;

	/* Amps limit (in dB relative to nominal 0) (float value) */
	LADSPA_Data limit = *(plugin_data->limit);

	/* Drive level (float value) */
	LADSPA_Data drive = *(plugin_data->drive);

	/* Low density colouration (float value) */
	LADSPA_Data locolour = *(plugin_data->locolour);

	/* High density colouration (float value) */
	LADSPA_Data hicolour = *(plugin_data->hicolour);

	/* Input L (array of floats of length sample_count) */
	LADSPA_Data *input1 = plugin_data->input1;

	/* Input R (array of floats of length sample_count) */
	LADSPA_Data *input2 = plugin_data->input2;

	/* Output L (array of floats of length sample_count) */
	LADSPA_Data *output1 = plugin_data->output1;

	/* Output R (array of floats of length sample_count) */
	LADSPA_Data *output2 = plugin_data->output2;

	LADSPA_Data gain_fix = pow(10, limit/20);
	
	for (pos = 0; pos < sample_count; pos++) {
	        /* Amplitude of signal */
	        LADSPA_Data amp1 = *(input1++) / gain_fix;
	        LADSPA_Data amp2 = *(input2++) / gain_fix;
	
	        *(output1++) = ((((1.0 - drive) * amp1) + (((1.0 / (1.0 + exp(-4.0 * amp1))) - 0.5) * 2.0 * drive) + (0.1 * locolour * sin(amp1 * 23) * amp1) + (0.1 * hicolour * cos(amp1 * 997) * amp1))) * gain_fix;
	        *(output2++) = ((((1.0 - drive) * amp2) + (((1.0 / (1.0 + exp(-4.0 * amp2))) - 0.5) * 2.0 * drive) + (0.1 * locolour * sin(amp2 * 23) * amp2) + (0.1 * hicolour * cos(amp2 * 997) * amp2))) * gain_fix;
	}
}

void _init() {
	char **port_names;
	LADSPA_PortDescriptor *port_descriptors;
	LADSPA_PortRangeHint *port_range_hints;

	mOverDrvDescriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (mOverDrvDescriptor) {
		mOverDrvDescriptor->UniqueID = 1182;
		mOverDrvDescriptor->Label = strdup("mOverDrv");
		mOverDrvDescriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		mOverDrvDescriptor->Name =
		 strdup("Mono Overdrive");
		mOverDrvDescriptor->Maker =
		 strdup("Steve Harris <steve@totl.net>");
		mOverDrvDescriptor->Copyright =
		 strdup("GPL");
		mOverDrvDescriptor->PortCount = 6;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(6,
		 sizeof(LADSPA_PortDescriptor));
		mOverDrvDescriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(6,
		 sizeof(LADSPA_PortRangeHint));
		mOverDrvDescriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(6, sizeof(char*));
		mOverDrvDescriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Amps limit (in dB relative to nominal 0) */
		port_descriptors[MOVERDRV_LIMIT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[MOVERDRV_LIMIT] =
		 strdup("Amps limit (in dB relative to nominal 0)");
		port_range_hints[MOVERDRV_LIMIT].HintDescriptor = 0;

		/* Parameters for Drive level */
		port_descriptors[MOVERDRV_DRIVE] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[MOVERDRV_DRIVE] =
		 strdup("Drive level");
		port_range_hints[MOVERDRV_DRIVE].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[MOVERDRV_DRIVE].LowerBound = 0;
		port_range_hints[MOVERDRV_DRIVE].UpperBound = +1;

		/* Parameters for Low density colouration */
		port_descriptors[MOVERDRV_LOCOLOUR] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[MOVERDRV_LOCOLOUR] =
		 strdup("Low density colouration");
		port_range_hints[MOVERDRV_LOCOLOUR].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[MOVERDRV_LOCOLOUR].LowerBound = 0;
		port_range_hints[MOVERDRV_LOCOLOUR].UpperBound = +1;

		/* Parameters for High density colouration */
		port_descriptors[MOVERDRV_HICOLOUR] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[MOVERDRV_HICOLOUR] =
		 strdup("High density colouration");
		port_range_hints[MOVERDRV_HICOLOUR].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[MOVERDRV_HICOLOUR].LowerBound = 0;
		port_range_hints[MOVERDRV_HICOLOUR].UpperBound = +1;

		/* Parameters for Input */
		port_descriptors[MOVERDRV_INPUT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[MOVERDRV_INPUT] =
		 strdup("Input");
		port_range_hints[MOVERDRV_INPUT].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[MOVERDRV_INPUT].LowerBound = -1;
		port_range_hints[MOVERDRV_INPUT].UpperBound = +1;

		/* Parameters for Output */
		port_descriptors[MOVERDRV_OUTPUT] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[MOVERDRV_OUTPUT] =
		 strdup("Output");
		port_range_hints[MOVERDRV_OUTPUT].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[MOVERDRV_OUTPUT].LowerBound = -1;
		port_range_hints[MOVERDRV_OUTPUT].UpperBound = +1;

		mOverDrvDescriptor->activate = NULL;
		mOverDrvDescriptor->cleanup = cleanupMOverDrv;
		mOverDrvDescriptor->connect_port = connectPortMOverDrv;
		mOverDrvDescriptor->deactivate = NULL;
		mOverDrvDescriptor->instantiate = instantiateMOverDrv;
		mOverDrvDescriptor->run = runMOverDrv;
		mOverDrvDescriptor->run_adding = NULL;
		mOverDrvDescriptor->set_run_adding_gain = NULL;
	}

	sOverDrvDescriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (sOverDrvDescriptor) {
		sOverDrvDescriptor->UniqueID = 1183;
		sOverDrvDescriptor->Label = strdup("sOverDrv");
		sOverDrvDescriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		sOverDrvDescriptor->Name =
		 strdup("Stereo Overdrive");
		sOverDrvDescriptor->Maker =
		 strdup("Steve Harris <steve@totl.net>");
		sOverDrvDescriptor->Copyright =
		 strdup("GPL");
		sOverDrvDescriptor->PortCount = 8;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(8,
		 sizeof(LADSPA_PortDescriptor));
		sOverDrvDescriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(8,
		 sizeof(LADSPA_PortRangeHint));
		sOverDrvDescriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(8, sizeof(char*));
		sOverDrvDescriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Amps limit (in dB relative to nominal 0) */
		port_descriptors[SOVERDRV_LIMIT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[SOVERDRV_LIMIT] =
		 strdup("Amps limit (in dB relative to nominal 0)");
		port_range_hints[SOVERDRV_LIMIT].HintDescriptor = 0;

		/* Parameters for Drive level */
		port_descriptors[SOVERDRV_DRIVE] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[SOVERDRV_DRIVE] =
		 strdup("Drive level");
		port_range_hints[SOVERDRV_DRIVE].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[SOVERDRV_DRIVE].LowerBound = 0;
		port_range_hints[SOVERDRV_DRIVE].UpperBound = 1;

		/* Parameters for Low density colouration */
		port_descriptors[SOVERDRV_LOCOLOUR] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[SOVERDRV_LOCOLOUR] =
		 strdup("Low density colouration");
		port_range_hints[SOVERDRV_LOCOLOUR].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW;
		port_range_hints[SOVERDRV_LOCOLOUR].LowerBound = 0;

		/* Parameters for High density colouration */
		port_descriptors[SOVERDRV_HICOLOUR] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[SOVERDRV_HICOLOUR] =
		 strdup("High density colouration");
		port_range_hints[SOVERDRV_HICOLOUR].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW;
		port_range_hints[SOVERDRV_HICOLOUR].LowerBound = 0;

		/* Parameters for Input L */
		port_descriptors[SOVERDRV_INPUT1] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[SOVERDRV_INPUT1] =
		 strdup("Input L");
		port_range_hints[SOVERDRV_INPUT1].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[SOVERDRV_INPUT1].LowerBound = -1;
		port_range_hints[SOVERDRV_INPUT1].UpperBound = +1;

		/* Parameters for Input R */
		port_descriptors[SOVERDRV_INPUT2] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[SOVERDRV_INPUT2] =
		 strdup("Input R");
		port_range_hints[SOVERDRV_INPUT2].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[SOVERDRV_INPUT2].LowerBound = -1;
		port_range_hints[SOVERDRV_INPUT2].UpperBound = +1;

		/* Parameters for Output L */
		port_descriptors[SOVERDRV_OUTPUT1] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[SOVERDRV_OUTPUT1] =
		 strdup("Output L");
		port_range_hints[SOVERDRV_OUTPUT1].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[SOVERDRV_OUTPUT1].LowerBound = -1;
		port_range_hints[SOVERDRV_OUTPUT1].UpperBound = +1;

		/* Parameters for Output R */
		port_descriptors[SOVERDRV_OUTPUT2] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[SOVERDRV_OUTPUT2] =
		 strdup("Output R");
		port_range_hints[SOVERDRV_OUTPUT2].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[SOVERDRV_OUTPUT2].LowerBound = -1;
		port_range_hints[SOVERDRV_OUTPUT2].UpperBound = +1;

		sOverDrvDescriptor->activate = NULL;
		sOverDrvDescriptor->cleanup = cleanupSOverDrv;
		sOverDrvDescriptor->connect_port = connectPortSOverDrv;
		sOverDrvDescriptor->deactivate = NULL;
		sOverDrvDescriptor->instantiate = instantiateSOverDrv;
		sOverDrvDescriptor->run = runSOverDrv;
		sOverDrvDescriptor->run_adding = NULL;
		sOverDrvDescriptor->set_run_adding_gain = NULL;
	}
}

void _fini() {
	int i;

	if (mOverDrvDescriptor) {
		free((char *)mOverDrvDescriptor->Label);
		free((char *)mOverDrvDescriptor->Name);
		free((char *)mOverDrvDescriptor->Maker);
		free((char *)mOverDrvDescriptor->Copyright);
		free((LADSPA_PortDescriptor *)mOverDrvDescriptor->PortDescriptors);
		for (i = 0; i < mOverDrvDescriptor->PortCount; i++)
			free((char *)(mOverDrvDescriptor->PortNames[i]));
		free((char **)mOverDrvDescriptor->PortNames);
		free((LADSPA_PortRangeHint *)mOverDrvDescriptor->PortRangeHints);
		free(mOverDrvDescriptor);
	}
	if (sOverDrvDescriptor) {
		free((char *)sOverDrvDescriptor->Label);
		free((char *)sOverDrvDescriptor->Name);
		free((char *)sOverDrvDescriptor->Maker);
		free((char *)sOverDrvDescriptor->Copyright);
		free((LADSPA_PortDescriptor *)sOverDrvDescriptor->PortDescriptors);
		for (i = 0; i < sOverDrvDescriptor->PortCount; i++)
			free((char *)(sOverDrvDescriptor->PortNames[i]));
		free((char **)sOverDrvDescriptor->PortNames);
		free((LADSPA_PortRangeHint *)sOverDrvDescriptor->PortRangeHints);
		free(sOverDrvDescriptor);
	}

}
